import java.io.*;
import java.sql.*;
import java.util.*;

/**
 * Programa interactivo en modo terminal que recibe instrucciones en SQL del usuario
 * y muestra los resultados de manera tabular.
 * <p>
 * Lee la configuracin de conexin a las bases de datos de un archivo.
 *
 * @author Gonzalo Mena Mendoza &lt;<a href="mailto:gonzalo@mena.com.mx">gonzalo@mena.com.mx</a>&gt;
 */
public class ConsultaDistribuida {

    final static int MAX_COL_ANCHO = 30;

    /**
     * "Dibuja" una tabla de texto simple que representa el contenido de un ResultSet.
     * Basado en Flanagan, D.: Java Examples in a Nutshell. O'Reilly (1997) 329-330
     */
    static void escribirResultado(OutputStream flujoSalida, ResultSet rs) throws SQLException {

        PrintWriter salida = new PrintWriter(flujoSalida);

        ResultSetMetaData metadatos = rs.getMetaData();         // informacion sobre las columnas

        // Propiedades de la tabla
        int      numCols        = metadatos.getColumnCount();   // nmero de columnas
        String[] colEtiquetas   = new String[numCols];          // etiqueta de cada columna
        int[]    colAnchos      = new int[numCols];             // ancho de cada columna
        int[]    colPos         = new int[numCols];             // posicin inicial de cada columna
        int      tabAncho;                                  // ancho total

        // Calcular ancho total, ancho y posicin de cada columna
        tabAncho = 1; // '|' al inicio.
        for(int i = 0; i < numCols; ++i) {                      // para cada columna
            colPos[i]       = tabAncho;                         //   posicin
            colEtiquetas[i] = metadatos.getColumnLabel(i + 1);  //   etiqueta
            int ancho = metadatos.getColumnDisplaySize(i + 1);  //   ancho en la BD
            if ((ancho == -1) || (ancho > MAX_COL_ANCHO)) {     //     por defecto
                ancho = MAX_COL_ANCHO;
            }
            int etiquetaAncho = colEtiquetas[i].length();
            if (etiquetaAncho > ancho) {
                ancho = etiquetaAncho;
            }
            colAnchos[i] = ancho + 1;                           // ancho de la columna
            tabAncho += colAnchos[i] + 2;                       // aumentar tamao total
        }

        // Lineas a usar
        StringBuffer divisor    = new StringBuffer(tabAncho);   // divisor horizontal
        StringBuffer lineaVacia = new StringBuffer(tabAncho);   // lnea en blanco
        StringBuffer linea;                                     // lnea actual
        for(int i = 0; i < tabAncho; ++i) {
            divisor   .insert(i, '-');
            lineaVacia.insert(i, " ");
        }

        for(int i = 0; i < numCols; ++i) {                      // marcas en el divisor
            divisor.setCharAt(colPos[i] - 1, '+');
        }
        divisor.setCharAt(tabAncho - 1, '+');

        // Encabezado con etiquetas
        salida.println(divisor);

        linea = new StringBuffer(lineaVacia.toString());
        linea.setCharAt(0, '|');
        for(int i = 0; i < numCols; ++i) {
            int pos = colPos[i] + 1 + ((colAnchos[i] - colEtiquetas[i].length()) / 2); // etiqueta centrada
            sobrescribir(linea, pos, colEtiquetas[i]);
            sobrescribir(linea, colPos[i] + colAnchos[i], " |");
        }

        salida.println(linea);
        salida.println(divisor);

        // Datos
        while(rs.next()) {
            linea = new StringBuffer(lineaVacia.toString());
            linea.setCharAt(0, '|');
            for(int i = 0; i < numCols; ++i) {
                Object valor = rs.getObject(i + 1);
                if (valor != null) {
                    sobrescribir(linea, colPos[i] + 1, valor.toString().trim());
                }
                sobrescribir(linea, colPos[i] + colAnchos[i], " |");
            }
            salida.println(linea);
        }

        // Final
        salida.println(divisor);
        salida.flush();
    }

    /**
     * Sobrescribe los caracteres del bfer a partir de la posicin dada con los de la cadena proporcionada.
     */
    static void sobrescribir(StringBuffer buf, int pos, String cad) {
        int cadLong = cad.length();         // longitud de la cadena
        int bufLong = buf.length();         // longitud del bfer
        if (pos + cadLong > bufLong)        // asegurar que quepa
            cadLong = bufLong - pos;
        for(int i = 0; i < cadLong; ++i)    // copiar cadena al bfer
            buf.setCharAt(pos + i, cad.charAt(i));
    }

    /**
     * Construye una instancia del programa.
     * Recibe un juego de propiedades con los datos de conexin local y remota.
     */
    ConsultaDistribuida(Properties props) {
        try {
            // ------------------------
            // propiedades de las conexiones
            // ------------------------

            // manejadores
            try {
                Class.forName(props.getProperty("local.manejador"));
                Class.forName(props.getProperty("remota.manejador"));
            }
            catch(ClassNotFoundException e) {
                throw new Exception("No se encontr el manejador de base de datos: " + e.getMessage(), e);
            }

            // obtener conexiones
            Connection conLocal  = null;
            Connection conRemota = null;
            try {
                conLocal = DriverManager.getConnection(
                    props.getProperty("local.url"),
                    props.getProperty("local.usuario"   , null),
                    props.getProperty("local.contrasena", null)
                );
                conRemota = DriverManager.getConnection(
                    props.getProperty("remota.url"),
                    props.getProperty("remota.usuario"   , null),
                    props.getProperty("remota.contrasena", null)
                );
            }
            catch(SQLException e) {
                throw new Exception("No se obtuvo conexin a la base de datos: " + e.getMessage(), e);
            }

            // tablas temporales locales?
            boolean tempLoc = (new Boolean(props.getProperty("local.temporal"))).booleanValue();

            // prefijo para diferenciar tabla remotas, p.e. "ext."
            String  prefijoRem = props.getProperty("remota.prefijo");

            System.out.print(
                "Local:  url=" + props.getProperty("local.url" ) + "\n" +
                "Remota: url=" + props.getProperty("remota.url") + "    prefijo=" + prefijoRem + "\n"
            );

            // ------------------------
            // procesar comandos de entrada hasta "salir"
            // ------------------------
            BufferedReader entrada = new BufferedReader(new InputStreamReader(System.in));
            while (true) {
                try {
                    System.out.print("\nsql> ");
                    String sql = entrada.readLine().trim();

                    if ((sql == null) || (sql.equals("salir")))
                        break;

                    if (sql.length() == 0)
                        continue;

                    // procesar comando, mostrar resultado si exite, o total de filas modificadas

                    long milis = System.currentTimeMillis(); // medir tiempo de ejecucn
                    ResultadoDistribuido res = new ResultadoDistribuido( // trabajo verdadero
                        conLocal, conRemota, sql, tempLoc, prefijoRem
                    );
                    milis = System.currentTimeMillis() - milis;

                    if (res.getResultSet() != null) {
                        System.out.println();
                        escribirResultado(System.out, res.getResultSet());
                    }
                    else {
                        System.out.println("\n" + res.getCuenta() + " filas modificadas.");
                    }

                    System.out.println("realizado en " + (milis / 1000.0) + " s.");
                    res.cerrar();

                }
                // sealar errores, normalmente de sintaxis, pero no salir de la aplicacin
                catch(SQLException e) {
                    System.err.println("Error: " + e.getMessage());
                }
                catch(IOException e) { // errores de entrada y salida
                    System.err.println("Error: " + e.getMessage());
                }
            }

            // importante cerrar conexiones
            conLocal.close();
            conRemota.close();
        }
        catch (Exception e) {
            System.err.println(e.getMessage());
        }
    }

    // ======================== principal ========================

    /**
     * Ejecuta una instancia del programa.
     * Lee la configuracin de un archivo.
     */
    public static void main(String[] args) throws Exception {
        System.out.print("\n" +
            "Consulta Distribuida\n" +
            "--------------------\n"
        );

        // ayuda
        if (args.length == 0) {
            System.out.print(
                "Utilizacin:\n" +
                "    java ConsultaDistribuida archivo_de_configuracin\n" +
                "\n" +
                "Ejemplo de propiedades en el archivo de configuracin:\n" +
                "    local.manejador=com.mysql.jdbc.Driver\n" +
                "    local.url=jdbc:mysql://localhost/amigos\n" +
                "    local.usuario=amigos\n" +
                "    local.contrasena=\n" +
                "    local.temporal=true\n" +
                "    remota.manejador=sun.jdbc.odbc.JdbcOdbcDriver\n" +
                "    remota.url=jdbc:odbc:empleados\n" +
                "    remota.usuario=\n" +
                "    remota.contrasena=\n" +
                "    remota.prefijo=ext\n"
            );
        }

        //  trabajo
        else {
            try {
                Properties props = new Properties();

                FileInputStream config = new FileInputStream(args[0]);
                props.load(config);
                config.close();

                new ConsultaDistribuida(props); // aqu esta la accin

            }
            catch (FileNotFoundException e) {
                System.err.println("No se encontr el archivo de configuracin.");
            }
            catch (IOException e) {
                System.err.println("No se pudo leer el archivo de configuracin.");
            }
        }
    }

}
