import java.sql.*;
import java.util.*;
import java.util.regex.*;

/**
 * Representa el resultado de una consulta distribuida sobre una base de datos local y una remota.
 * Los clientes acceden a las tablas remotas diferencindolas de los locales
 * mediante un prefijo definido por el usuario.
 * La presencia de tablas remotas es completamente transparente para el usuario,
 * quien las percibe como si se tratase de tablas de otro esquema o catlogo local.
 *
 * @author Gonzalo Mena Mendoza &lt;<a href="mailto:gonzalo@mena.com.mx">gonzalo@mena.com.mx</a>&gt;
 */
public class ResultadoDistribuido {

    // ======================== constantes ========================

    final static String[] ARR_VACIO = {};

    final static String RE_NOMBRE           = "[\\p{Alpha}_](?:[\\p{Alnum}_])*";
    final static String RE_TABLA_LOCAL      = RE_NOMBRE;
    final static String RE_COLUMNA_LOCAL    = RE_NOMBRE + "\\." + RE_NOMBRE;

    final static String RE_SEPARADOR        = "\\s*,\\s*";
    final static String RE_CONJUNCION       = "(?i)\\s+AND\\s+";

    // Patrn general de consulta, sin reuniones externas
    static Pattern PATRON_CONSULTA = Pattern.compile(
        "SELECT\\s+(.+)\\s+FROM\\s+(.+?)(\\s+WHERE\\s+(.+))?;?",
        Pattern.CASE_INSENSITIVE | Pattern.MULTILINE | Pattern.DOTALL
    );

    // Patrn de condiciones: expresin operador expresin
    static Pattern PATRON_CONDICION = Pattern.compile(
        "(.+?)\\s+(=|<>|<|>|LIKE|NOT\\s+LIKE|IS|IS\\s+NOT)\\s+(.+?)",
        Pattern.CASE_INSENSITIVE | Pattern.MULTILINE | Pattern.DOTALL
    );

    // Patrn creacin, retiro, actualizacin e insercin en tablas remotas
    static Pattern PATRON_DIRECTA_REMOTA = Pattern.compile(
        "(CREATE\\s+TABLE|DROP\\s+TABLE|UPDATE|INSERT\\s+INTO)\\s+" + RE_NOMBRE + "\\.(.+)",
        Pattern.CASE_INSENSITIVE | Pattern.MULTILINE | Pattern.DOTALL
    );

    final static String TABLA_TEMPORAL      = "TEMP";

    final static boolean DEPURAR            = true;

    // ======================== propiedades ========================

    private Connection conLoc       = null;
    private Connection conRem       = null;
    private String     sql          = null;
    private boolean    tempLoc      = false;
    private String     prefijoRem   = null;
    private int        cuenta       = -1;   // total de filas modificadas
    private ResultSet  rs           = null; // conjunto de resultados finales

    // ======================== principal ========================

    /**
     * Construye y procesa un nuevo resultado distribuido
     * sobre una base de datos local y una remota
     * a partir de la instruccion en SQL proporcionada.
     */
    ResultadoDistribuido(Connection conLoc, Connection conRem, String sql, boolean tempLoc, String prefijoRem)
      throws SQLException
    {
        this.conLoc     = conLoc;
        this.conRem     = conRem;
        this.sql        = sql;
        this.tempLoc    = tempLoc;
        this.prefijoRem = prefijoRem;

        procesar();
    }

    ResultSet getResultSet() {
        return rs;
    }

    int getCuenta() {
        return cuenta;
    }

    void cerrar() throws SQLException {
        if ((rs != null) && (rs.getStatement() != null)) {
            rs.getStatement().close();
        }
    }

    /**
     * Procesa la instruccin de este resultado.
     * Compara dicho SQL con cada uno  de los patrones considerados
     * y la descompone en diferentes instrucciones segn sea necesario.
     */
    void procesar() throws SQLException {
        Matcher m = null;

        // SELECT ----------------------------------------
        m = PATRON_CONSULTA.matcher(sql);
        if (m.matches()) {

            // Listas de columnas, tablas y condiciones separadas por comas o conjunciones
            String listaColumnas = m.group(1);
            String listaTablas   = m.group(2);
            String listaConds    = (m.groupCount() > 2) ? m.group(4) : null;

            // Arreglos de columnas, tablas y condiciones generales
            String[] columnas = listaColumnas.split(RE_SEPARADOR, -1);
            String[] tablas   = listaTablas  .split(RE_SEPARADOR, -1);
            String[] conds    = (listaConds == null) ? ARR_VACIO
                              : listaConds.split(RE_CONJUNCION, -2);

            // Conjuntos de tablas, columnas y condiciones remotas
            Set tablasRem   = new LinkedHashSet();
            Set columnasRem = new LinkedHashSet();
            Set condsRem    = new LinkedHashSet();

            // Conjuntos de tablas, columnas y condiciones locales
            Set tablasLoc   = new LinkedHashSet();
            Set columnasLoc = new LinkedHashSet();
            Set condsLoc    = new LinkedHashSet();

            // Dividir tablas en remotas y locales
            for (int i = 0; i < tablas.length; ++i) {
                if (esTablaRemota(tablas[i])) {
                    tablasRem.add(tablaLocal(tablas[i]));
                    tablasLoc.add(TABLA_TEMPORAL);
                }
                else {
                    tablasLoc.add(tablas[i]);
                }
            }

            // Dividir columnas en remotas y locales
            for (int i = 0; i < columnas.length; ++i) {
                if (esColumnaRemota(columnas[i])) {
                    columnasRem.add(columnaLocal( columnas[i]));
                    columnasLoc.add(columnaTemp(columnas[i]) + " AS " + columnaSola(columnas[i]));
                }
                else {
                    columnasLoc.add(columnas[i]);
                }
            }

            // Agrupar las condiciones en remotas y locales
            for (int i = 0; i < conds.length; ++i) {
                Matcher mc = PATRON_CONDICION.matcher(conds[i]);

                if (mc.matches()) {
                    String izq = mc.group(1); // expresin izquierda, columna o escalar
                    String op  = mc.group(2); // operador
                    String der = mc.group(3); // expresin derecha, columna o escalar

                    // dos columnas remotas
                    if (esColumnaRemota(izq)  && esColumnaRemota(der)) {
                        condsRem.add(columnaLocal(izq) + " " + op + " " + columnaLocal(der));
                    }

                    // columna remota y escalar
                    else if (esColumnaRemota(izq)  && esEscalar(der)) {
                        condsRem.add(columnaLocal(izq) + " " + op + " " + der);
                    }
                    else if (esEscalar(izq) && esColumnaRemota(der)) {
                        condsRem.add(izq + " " + op + " " + columnaLocal(der));
                    }

                    // columna remota y columna local: aade remota a resultado para comparar localmente
                    else if (esColumnaRemota(izq)) {
                        columnasRem.add(columnaLocal(izq));
                        condsLoc.add(columnaTemp(izq) + " " + op + " " + der);
                    }
                    else if (esColumnaRemota(der)) {
                        columnasRem.add(columnaLocal(der));
                        condsLoc.add(izq + " " + op + " " + columnaTemp(der));
                    }

                    // dos columnas o expresiones locales
                    else {
                        condsLoc.add(izq + " " + op + " " + der);
                    }
                }
            }

            String sqlRem = construirConsulta(tablasRem, columnasRem, condsRem);
            String sqlLoc = construirConsulta(tablasLoc, columnasLoc, condsLoc);

            Statement stLoc = conLoc.createStatement();

            // Ejecutar consulta remota
            if (sqlRem != null) {
                if (DEPURAR) System.out.println("\n---- remota ----\n\n" + sqlRem);

                Statement stRem = conRem.createStatement();
                ResultSet rsRem = stRem.executeQuery(sqlRem);

                // Regresar resultados directamente si no hay consulta local
                if ((tablasLoc.size() == 1) && tablasLoc.contains(TABLA_TEMPORAL)) {
                    rs = rsRem;
                    return;
                }

                // Retirar tabla temporal existente
                try {
                    String ddl = "DROP TABLE " + TABLA_TEMPORAL;
                    if (DEPURAR) System.out.println("\n---- local ----\n\n" + ddl);
                    stLoc.executeUpdate(ddl);
                }
                catch(SQLException e) {}

                // Crear nueva tabla temporal local en base a resultados remotos
                String ddl = creacionTablaTemporal(rsRem, tempLoc);
                if (DEPURAR) System.out.println("\n" + ddl);
                stLoc.executeUpdate(ddl);

                // Llenar tabla temporal local con resultados remotos
                String sqlAux = llenarTablaTemporal(conLoc, rsRem);
                if (DEPURAR) System.out.println("\n" + sqlAux);

                rsRem.close();
                stRem.close();
            }

            // Ejecutar consulta local que puede incluir a la tabla temporal con resultados remotos
            if (DEPURAR) System.out.println("\n---- local ----\n\n"  + sqlLoc);
            rs      = stLoc.executeQuery(sqlLoc);
            cuenta  = stLoc.getUpdateCount();

            return;
        }

        // CREATE, DROP, INSERT, UPDATE ----------------------------------------
        m = PATRON_DIRECTA_REMOTA.matcher(sql);
        if (m.matches()) {
            String sqlRem =  m.group(1) + " " + m.group(2);
            if (DEPURAR) System.out.println("\n---- remota ----\n\n" + sqlRem);

            Statement stRem = conRem.createStatement();
            stRem.execute(sqlRem);
            cuenta = stRem.getUpdateCount();
            stRem.close();

            return;
        }

        // Patrn no considerado ----------------------------------------
        if (DEPURAR) System.out.println("\n---- local, otro ----");
        Statement stLoc = conLoc.createStatement();
        stLoc.execute(sql);
        rs      = stLoc.getResultSet();
        cuenta  = stLoc.getUpdateCount();
    }

    // ======================== herramientas ========================

    boolean esEscalar(String s) {
        return !esColumnaRemota(s) && !esColumnaLocal(s);
    }

    boolean esTablaRemota(String s) {
        return s.trim().matches(prefijoRem + "\\." + RE_TABLA_LOCAL);
    }

    static boolean esTablaLocal(String s) {
        return s.trim().matches(RE_TABLA_LOCAL);
    }

    boolean esColumnaRemota(String s) {
        return s.trim().matches(prefijoRem + "\\." + RE_COLUMNA_LOCAL);
    }

    static boolean esColumnaLocal(String s) {
        return s.trim().matches(RE_COLUMNA_LOCAL);
    }

    static String tablaLocal(String s) {
        return s.trim().replaceAll("^" + RE_NOMBRE + "\\.", "");
    }

    static String columnaSola(String s) {
        return s.trim().replaceAll(".*\\.([^\\.]+)", "$1");
    }

    static String columnaLocal(String s) {
        return s.trim().replaceAll("^" + RE_NOMBRE + "\\.", "");
    }

    static String columnaUnica(String s) {
        return s.trim().replaceAll("\\.", "_");
    }

    static String columnaTemp(String s) {
        return TABLA_TEMPORAL + "." + columnaUnica(columnaLocal(s));
    }

    /**
     * Regresa la instruccion en SQL necesaria para realizar una consulta
     * en base a los conjuntos de tablas, columnas y condiciones proporcionadas.
     * Estas ultimas son opcionales.
     */
    static String construirConsulta(Set tablas, Set columnas, Set conds) {
        if ((tablas.size() < 1) || (columnas.size() < 1))
            return null;

        StringBuffer sql = new StringBuffer(
            "SELECT " + juntar(columnas, ", ") + " \n" +
            "  FROM " + juntar(tablas,   ", ")
        );
        if (conds.size() > 0)
            sql.append("\n WHERE " + juntar(conds, "\n   AND "));

        return sql.toString();
    }

    /**
     * Regresa la instruccion en SQL necesaria para crear una tabla
     * que contenga todas las columnas del conjunto de resultados proporcionado.
     * Considera el nombre de la tabla origen ya que en los resultados varias columnas pueden tener el mismo nombre.
     */
    static String creacionTablaTemporal(ResultSet rs, boolean temp) throws SQLException {
        List defs = new ArrayList();

        ResultSetMetaData metadatos = rs.getMetaData();
        int numCols = metadatos.getColumnCount();
        for(int i = 1; i <= numCols; ++i) {
            // determinar tipo de dato
            String tipo     = metadatos.getColumnTypeName(i);
            int    tamano   = metadatos.getColumnDisplaySize(i);
            if (tipo.equals("NUMBER")) {
                tipo = "int";
                tamano = -1;
            }
            // cada definicion de columna es de la forma "Tabla_columna TIPO(TAMANO)"
            defs.add(
                columnaUnica(metadatos.getTableName(i) + "." + metadatos.getColumnName(i)) +
                " " + tipo + ((tamano > -1) ? ("(" + tamano + ")") : "")
            );
        }

        return (
            "CREATE " + (temp ? "TEMPORARY " : "") + "TABLE " + TABLA_TEMPORAL + " (" + juntar(defs, ", ") + ")"
        );
    }

    /**
     * Inserta un fila en la tabla temporal por cada fila del conjunto de resultados proporcionado.
     * Regresa la instruccion parametrizada en SQL para realizar cada insercin.
     */
    static String llenarTablaTemporal(Connection con, ResultSet rs) throws SQLException {
        List cols = new ArrayList();
        List esps = new ArrayList();

        ResultSetMetaData metadatos = rs.getMetaData();
        int numCols = metadatos.getColumnCount();
        for(int i = 1; i <= numCols; ++i) {
            cols.add(columnaUnica(metadatos.getTableName(i) + "." + metadatos.getColumnName(i)));
            esps.add("?");
        }

        String sql = (
            "INSERT INTO " + TABLA_TEMPORAL + " (" + juntar(cols, ", ") + ") VALUES (" + juntar(esps, ", ") + ")"
        );

        PreparedStatement ps = con.prepareStatement(sql);
        while (rs.next()) {
            for(int i = 1; i <= numCols; ++i) {
                ps.setObject(i, rs.getObject(i));
            }
            ps.execute();
        }

        return sql;
    }

    /**
     * Regresa una cadena con todos los elementos de la coleccion separados por el texto separador proporcionado.
     */
    static String juntar(Collection col, String sep) {
        StringBuffer aux = null;
        for (Iterator i = col.iterator(); i.hasNext();) {
            if (null == aux)
                aux = new StringBuffer(i.next().toString());
            else
                aux.append(sep + i.next());
        }

        return (aux != null) ? aux.toString() : null;
    }

}
